@@ -1,10 +1,10 @@
 // Copyright (C) 2016 The Qt Company Ltd.
 // SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
-#include <qcameradevice.h>
+#include <qgstreamercamera_p.h>
 
-#include "qgstreamercamera_p.h"
-#include "qgstreamerimagecapture_p.h"
+#include <qcameradevice.h>
+#include <qgst_debug_p.h>
 #include <qgstreamervideodevices_p.h>
 #include <qgstreamerintegration_p.h>
 #include <qmediacapturesession.h>
@@ -20,19 +20,19 @@ QT_BEGIN_NAMESPACE
 
 QMaybe<QPlatformCamera *> QGstreamerCamera::create(QCamera *camera)
 {
-    QGstElement videotestsrc("videotestsrc");
+    QGstElement videotestsrc = QGstElement::createFromFactory("videotestsrc");
     if (!videotestsrc)
         return errorMessageCannotFindElement("videotestsrc");
 
-    QGstElement capsFilter("capsfilter", "videoCapsFilter");
+    QGstElement capsFilter = QGstElement::createFromFactory("capsfilter", "videoCapsFilter");
     if (!capsFilter)
         return errorMessageCannotFindElement("capsfilter");
 
-    QGstElement videoconvert("videoconvert", "videoConvert");
+    QGstElement videoconvert = QGstElement::createFromFactory("videoconvert", "videoConvert");
     if (!videoconvert)
         return errorMessageCannotFindElement("videoconvert");
 
-    QGstElement videoscale("videoscale", "videoScale");
+    QGstElement videoscale = QGstElement::createFromFactory("videoscale", "videoScale");
     if (!videoscale)
         return errorMessageCannotFindElement("videoscale");
 
@@ -48,20 +48,15 @@ QGstreamerCamera::QGstreamerCamera(QGstElement videotestsrc, QGstElement capsFil
       gstVideoConvert(std::move(videoconvert)),
       gstVideoScale(std::move(videoscale))
 {
-    gstDecode = QGstElement("identity");
-    gstCameraBin = QGstBin("camerabin");
+    gstDecode = QGstElement::createFromFactory("identity");
+    gstCameraBin = QGstBin::create("camerabin");
     gstCameraBin.add(gstCamera, gstCapsFilter, gstDecode, gstVideoConvert, gstVideoScale);
-    gstCamera.link(gstCapsFilter, gstDecode, gstVideoConvert, gstVideoScale);
+    qLinkGstElements(gstCamera, gstCapsFilter, gstDecode, gstVideoConvert, gstVideoScale);
     gstCameraBin.addGhostPad(gstVideoScale, "src");
 }
 
 QGstreamerCamera::~QGstreamerCamera()
 {
-#if QT_CONFIG(linux_v4l)
-    if (v4l2FileDescriptor >= 0)
-        qt_safe_close(v4l2FileDescriptor);
-    v4l2FileDescriptor = -1;
-#endif
     gstCameraBin.setStateSync(GST_STATE_NULL);
 }
 
@@ -91,47 +86,37 @@ void QGstreamerCamera::setCamera(const QCameraDevice &camera)
 
     QGstElement gstNewCamera;
     if (camera.isNull()) {
-        gstNewCamera = QGstElement("videotestsrc");
+        gstNewCamera = QGstElement::createFromFactory("videotestsrc");
     } else {
         auto *integration = static_cast<QGstreamerIntegration *>(QGstreamerIntegration::instance());
         auto *device = integration->videoDevice(camera.id());
-        gstNewCamera = gst_device_create_element(device, "camerasrc");
+        gstNewCamera = QGstElement(gst_device_create_element(device, "camerasrc"));
         if (QGstStructure properties = gst_device_get_properties(device); !properties.isNull()) {
             if (properties.name() == "v4l2deviceprovider")
-                m_v4l2Device = QString::fromUtf8(properties["device.path"].toString());
+                m_v4l2DevicePath = QString::fromUtf8(properties["device.path"].toString());
             properties.free();
         }
     }
 
     QCameraFormat f = findBestCameraFormat(camera);
     auto caps = QGstCaps::fromCameraFormat(f);
-    auto gstNewDecode = QGstElement(f.pixelFormat() == QVideoFrameFormat::Format_Jpeg ? "jpegdec" : "identity");
+    auto gstNewDecode = QGstElement::createFromFactory(
+            f.pixelFormat() == QVideoFrameFormat::Format_Jpeg ? "jpegdec" : "identity");
 
-    gstCamera.unlink(gstCapsFilter);
-    gstCapsFilter.unlink(gstDecode);
-    gstDecode.unlink(gstVideoConvert);
-
-    gstCameraBin.remove(gstCamera);
-    gstCameraBin.remove(gstDecode);
-
-    gstCamera.setStateSync(GST_STATE_NULL);
-    gstDecode.setStateSync(GST_STATE_NULL);
+    qUnlinkGstElements(gstCamera, gstCapsFilter, gstDecode, gstVideoConvert);
+    gstCameraBin.stopAndRemoveElements(gstCamera, gstDecode);
 
     gstCapsFilter.set("caps", caps);
 
     gstCameraBin.add(gstNewCamera, gstNewDecode);
-
-    gstNewDecode.link(gstVideoConvert);
-    gstCapsFilter.link(gstNewDecode);
-
-    if (!gstNewCamera.link(gstCapsFilter))
-        qWarning() << "linking camera failed" << gstCamera.name() << caps.toString();
+    qLinkGstElements(gstNewCamera, gstCapsFilter, gstNewDecode, gstVideoConvert);
 
     // Start sending frames once pipeline is linked
-    // FIXME: put camera to READY state before linking to decoder as in the NULL state it does not know its true caps
-    gstCapsFilter.syncStateWithParent();
-    gstNewDecode.syncStateWithParent();
-    gstNewCamera.syncStateWithParent();
+    // FIXME: put camera to READY state before linking to decoder as in the NULL state it does not
+    // know its true caps
+
+    gstCameraBin.syncChildrenState();
+
 
     gstCamera = gstNewCamera;
     gstDecode = gstNewDecode;
@@ -150,25 +135,19 @@ bool QGstreamerCamera::setCameraFormat(const QCameraFormat &format)
 
     auto caps = QGstCaps::fromCameraFormat(f);
 
-    auto newGstDecode = QGstElement(f.pixelFormat() == QVideoFrameFormat::Format_Jpeg ? "jpegdec" : "identity");
+    auto newGstDecode = QGstElement::createFromFactory(
+            f.pixelFormat() == QVideoFrameFormat::Format_Jpeg ? "jpegdec" : "identity");
     gstCameraBin.add(newGstDecode);
     newGstDecode.syncStateWithParent();
 
-    gstCamera.staticPad("src").doInIdleProbe([&](){
-        gstCamera.unlink(gstCapsFilter);
-        gstCapsFilter.unlink(gstDecode);
-        gstDecode.unlink(gstVideoConvert);
+    gstCamera.staticPad("src").doInIdleProbe([&]() {
+        qUnlinkGstElements(gstCamera, gstCapsFilter, gstDecode, gstVideoConvert);
 
         gstCapsFilter.set("caps", caps);
-
-        newGstDecode.link(gstVideoConvert);
-        gstCapsFilter.link(newGstDecode);
-        if (!gstCamera.link(gstCapsFilter))
-            qWarning() << "linking filtered camera to decoder failed" << gstCamera.name() << caps.toString();
+        qLinkGstElements(gstCamera, gstCapsFilter, newGstDecode, gstVideoConvert);
     });
 
-    gstCameraBin.remove(gstDecode);
-    gstDecode.setStateSync(GST_STATE_NULL);
+    gstCameraBin.stopAndRemoveElements(gstDecode);
 
     gstDecode = newGstDecode;
 
@@ -609,83 +588,78 @@ void QGstreamerCamera::setColorTemperature(int temperature)
 }
 
 #if QT_CONFIG(linux_v4l)
+bool QGstreamerCamera::isV4L2Camera() const
+{
+    return !m_v4l2DevicePath.isEmpty();
+}
+
 void QGstreamerCamera::initV4L2Controls()
 {
     v4l2AutoWhiteBalanceSupported = false;
     v4l2ColorTemperatureSupported = false;
-    QCamera::Features features;
+    QCamera::Features features{};
+
+    Q_ASSERT(!m_v4l2DevicePath.isEmpty());
 
 
-    const QString deviceName = v4l2Device();
-    Q_ASSERT(!deviceName.isEmpty());
+    withV4L2DeviceFileDescriptor([&](int fd) {
+        struct v4l2_queryctrl queryControl = {};
+        queryControl.id = V4L2_CID_AUTO_WHITE_BALANCE;
 
-    v4l2FileDescriptor = qt_safe_open(deviceName.toLocal8Bit().constData(), O_RDONLY);
-    if (v4l2FileDescriptor == -1) {
-        qWarning() << "Unable to open the camera" << deviceName
-                   << "for read to query the parameter info:" << qt_error_string(errno);
-        return;
-    }
-
-    struct v4l2_queryctrl queryControl;
-    ::memset(&queryControl, 0, sizeof(queryControl));
-    queryControl.id = V4L2_CID_AUTO_WHITE_BALANCE;
-
-    if (::ioctl(v4l2FileDescriptor, VIDIOC_QUERYCTRL, &queryControl) == 0) {
-        v4l2AutoWhiteBalanceSupported = true;
-        setV4L2Parameter(V4L2_CID_AUTO_WHITE_BALANCE, true);
-    }
-
-    ::memset(&queryControl, 0, sizeof(queryControl));
-    queryControl.id = V4L2_CID_WHITE_BALANCE_TEMPERATURE;
-    if (::ioctl(v4l2FileDescriptor, VIDIOC_QUERYCTRL, &queryControl) == 0) {
-        v4l2MinColorTemp = queryControl.minimum;
-        v4l2MaxColorTemp = queryControl.maximum;
-        v4l2ColorTemperatureSupported = true;
-        features |= QCamera::Feature::ColorTemperature;
-    }
-
-    ::memset(&queryControl, 0, sizeof(queryControl));
-    queryControl.id = V4L2_CID_EXPOSURE_AUTO;
-    if (::ioctl(v4l2FileDescriptor, VIDIOC_QUERYCTRL, &queryControl) == 0) {
-        v4l2AutoExposureSupported = true;
-    }
-
-    ::memset(&queryControl, 0, sizeof(queryControl));
-    queryControl.id = V4L2_CID_EXPOSURE_ABSOLUTE;
-    if (::ioctl(v4l2FileDescriptor, VIDIOC_QUERYCTRL, &queryControl) == 0) {
-        v4l2ManualExposureSupported = true;
-        v4l2MinExposure = queryControl.minimum;
-        v4l2MaxExposure = queryControl.maximum;
-        features |= QCamera::Feature::ManualExposureTime;
-    }
-
-    ::memset(&queryControl, 0, sizeof(queryControl));
-    queryControl.id = V4L2_CID_AUTO_EXPOSURE_BIAS;
-    if (::ioctl(v4l2FileDescriptor, VIDIOC_QUERYCTRL, &queryControl) == 0) {
-        v4l2MinExposureAdjustment = queryControl.minimum;
-        v4l2MaxExposureAdjustment = queryControl.maximum;
-        features |= QCamera::Feature::ExposureCompensation;
-    }
-
-    ::memset(&queryControl, 0, sizeof(queryControl));
-    queryControl.id = V4L2_CID_ISO_SENSITIVITY_AUTO;
-    if (::ioctl(v4l2FileDescriptor, VIDIOC_QUERYCTRL, &queryControl) == 0) {
-        queryControl.id = V4L2_CID_ISO_SENSITIVITY;
-        if (::ioctl(v4l2FileDescriptor, VIDIOC_QUERYCTRL, &queryControl) == 0) {
-            features |= QCamera::Feature::IsoSensitivity;
-            minIsoChanged(queryControl.minimum);
-            maxIsoChanged(queryControl.minimum);
+        if (::ioctl(fd, VIDIOC_QUERYCTRL, &queryControl) == 0) {
+            v4l2AutoWhiteBalanceSupported = true;
+            setV4L2Parameter(V4L2_CID_AUTO_WHITE_BALANCE, true);
         }
-    }
+
+        queryControl = {};
+        queryControl.id = V4L2_CID_WHITE_BALANCE_TEMPERATURE;
+        if (::ioctl(fd, VIDIOC_QUERYCTRL, &queryControl) == 0) {
+            v4l2MinColorTemp = queryControl.minimum;
+            v4l2MaxColorTemp = queryControl.maximum;
+            v4l2ColorTemperatureSupported = true;
+            features |= QCamera::Feature::ColorTemperature;
+        }
+
+        queryControl = {};
+        queryControl.id = V4L2_CID_EXPOSURE_AUTO;
+        if (::ioctl(fd, VIDIOC_QUERYCTRL, &queryControl) == 0) {
+            v4l2AutoExposureSupported = true;
+        }
+
+        queryControl = {};
+        queryControl.id = V4L2_CID_EXPOSURE_ABSOLUTE;
+        if (::ioctl(fd, VIDIOC_QUERYCTRL, &queryControl) == 0) {
+            v4l2ManualExposureSupported = true;
+            v4l2MinExposure = queryControl.minimum;
+            v4l2MaxExposure = queryControl.maximum;
+            features |= QCamera::Feature::ManualExposureTime;
+        }
+
+        queryControl = {};
+        queryControl.id = V4L2_CID_AUTO_EXPOSURE_BIAS;
+        if (::ioctl(fd, VIDIOC_QUERYCTRL, &queryControl) == 0) {
+            v4l2MinExposureAdjustment = queryControl.minimum;
+            v4l2MaxExposureAdjustment = queryControl.maximum;
+            features |= QCamera::Feature::ExposureCompensation;
+        }
+
+        queryControl = {};
+        queryControl.id = V4L2_CID_ISO_SENSITIVITY_AUTO;
+        if (::ioctl(fd, VIDIOC_QUERYCTRL, &queryControl) == 0) {
+            queryControl.id = V4L2_CID_ISO_SENSITIVITY;
+            if (::ioctl(fd, VIDIOC_QUERYCTRL, &queryControl) == 0) {
+                features |= QCamera::Feature::IsoSensitivity;
+                minIsoChanged(queryControl.minimum);
+                maxIsoChanged(queryControl.minimum);
+            }
+        }
+    });
 
     supportedFeaturesChanged(features);
 }
 
 int QGstreamerCamera::setV4L2ColorTemperature(int temperature)
 {
-    struct v4l2_control control;
-    ::memset(&control, 0, sizeof(control));
-
     if (v4l2AutoWhiteBalanceSupported) {
         setV4L2Parameter(V4L2_CID_AUTO_WHITE_BALANCE, temperature == 0 ? true : false);
     } else if (temperature == 0) {
@@ -705,22 +679,28 @@ int QGstreamerCamera::setV4L2ColorTemperature(int temperature)
 
 bool QGstreamerCamera::setV4L2Parameter(quint32 id, qint32 value)
 {
-    struct v4l2_control control{id, value};
-    if (::ioctl(v4l2FileDescriptor, VIDIOC_S_CTRL, &control) != 0) {
-        qWarning() << "Unable to set the V4L2 Parameter" << Qt::hex << id << "to" << value << qt_error_string(errno);
-        return false;
-    }
-    return true;
+    return withV4L2DeviceFileDescriptor([&](int fd) {
+        v4l2_control control{ id, value };
+        if (::ioctl(fd, VIDIOC_S_CTRL, &control) != 0) {
+            qWarning() << "Unable to set the V4L2 Parameter" << Qt::hex << id << "to" << value
+                       << qt_error_string(errno);
+            return false;
+        }
+        return true;
+    });
 }
 
 int QGstreamerCamera::getV4L2Parameter(quint32 id) const
 {
-    struct v4l2_control control{id, 0};
-    if (::ioctl(v4l2FileDescriptor, VIDIOC_G_CTRL, &control) != 0) {
-        qWarning() << "Unable to get the V4L2 Parameter" << Qt::hex << id << qt_error_string(errno);
-        return 0;
-    }
-    return control.value;
+    return withV4L2DeviceFileDescriptor([&](int fd) {
+        v4l2_control control{ id, 0 };
+        if (::ioctl(fd, VIDIOC_G_CTRL, &control) != 0) {
+            qWarning() << "Unable to get the V4L2 Parameter" << Qt::hex << id
+                       << qt_error_string(errno);
+            return 0;
+        }
+        return control.value;
+    });
 }
 
 #endif
